import re
from typing import Final

from markdown import Markdown
from markdown.extensions import Extension
from markdown.extensions.admonition import AdmonitionProcessor
from markdown.preprocessors import Preprocessor

from .logging import get_logger

log = get_logger(__name__)

# TODO: implement ways to actually use the information ...


class AdmotionCellDataPreprocessor(Preprocessor):

    CELL_REGEX: Final = re.compile(r"^::: \{\.cell .*}\s*$")
    CELL_END: Final = re.compile(r"^:::$")
    CELL_ELEM_REGEX: Final = re.compile(r"^::: \{(.cell-\w+) (\.cell-[\w-]+)\}$")
    CODEBLOCK_REGEX: Final = re.compile(r"^```{\.(\w+) .*}")

    # https://squidfunk.github.io/mkdocs-material/reference/admonitions/#supported-types
    TYPE_MAPPING: Final = {
        ".cell-output-stdout": '!!! note "output"',
        ".cell-output-stderr": '!!! warning "stderr"',
        ".cell-output-error": '!!! danger "error"',
        ".cell-output-display": '!!! note "Figure"',
    }

    def run(self, lines):
        log.info(f"Running {self}")
        outs = [self._process_line(x) for x in lines]
        log.debug(f"Removing {sum(1 for x in outs if x is None)} lines")
        out = [x for x in outs if x is not None]

        return out

    def _process_line(self, line):
        if sr := self.CELL_REGEX.search(line):
            log.debug(f"Matched Cell start: {line}")
            out = "\n\n"

        elif sr := self.CELL_END.search(line):
            log.debug(f"Matched Cell end: {line}")
            out = "\n\n"

        elif sr := self.CELL_ELEM_REGEX.search(line):
            log.debug(f"Matched Cell element: {line}")
            _, output_type = sr.groups()
            out = self.TYPE_MAPPING[output_type]

        elif sr := self.CODEBLOCK_REGEX.search(line):
            log.debug(f"Matched codeblock: {line}")
            lang = sr.groups(1)
            out = f"```{lang}"
        else:
            out = line

        if line != out:
            log.debug(f"Transformed {line} -> {out}")
        return out


# TODO consider implemeting this ...
# class CellBLockProcessor(BlockProcessor):


class QuartoCellDataExtension(Extension):
    def extendMarkdown(
        self, md: Markdown
    ) -> None:  # noqa: N802 (casing: parent method's name)
        """Register the extension.

        Adds an instance of the Processor to the Markdown instance.
            md: A `markdown.Markdown` instance.
        """
        md.registerExtension(self)
        md.preprocessors.register(
            AdmotionCellDataPreprocessor(),
            name="QuatoCellData",
            priority=102,  # Right before Admonition
        )
        md.parser.blockprocessors.register(
            AdmonitionProcessor(md.parser), "admonition", 105
        )



import re
import shutil
import subprocess
import warnings
from pathlib import Path

from mkdocs.config import config_options
from mkdocs.plugins import BasePlugin

from mkquartodocs.extension import QuartoCellDataExtension

from .context import DirWatcherContext
from .logging import get_logger

log = get_logger(__name__)


def _delete_file(path):
    """Delete a file or an empty directory."""
    path = Path(path)
    if path.is_file():
        path.unlink()
    elif path.is_dir():
        path.rmdir()


class MkDocstringPlugin(BasePlugin):
    config_scheme = (
        ("quarto_path", config_options.Type(Path)),
        ("ignore", config_options.Type(str)),
        ("keep_output", config_options.Type(bool, default=False)),
    )

    def on_config(self, config, **kwargs):
        passed_path = self.config["quarto_path"]
        quarto = shutil.which(passed_path if passed_path else "quarto")
        self.config["quarto_path"] = quarto
        # self.ignores = [re.compile(x) for x in self.config["ignore"]]

        if self.config["ignore"]:
            self.ignores = [re.compile(self.config["ignore"])]
        else:
            self.ignores = []
        self.exit_action = _delete_file if not self.config["keep_output"] else None

        self.extension = QuartoCellDataExtension()
        config["markdown_extensions"].append(self.extension)
        return config

    def _filter_ignores(self, paths):
        out = []
        for x in paths:
            if not any(re.fullmatch(pattern, x) for pattern in self.ignores):
                out.append(x)

        return out

    def on_pre_build(self, config):
        quarto = self.config["quarto_path"]
        docs_dir = config["docs_dir"]

        quarto_docs = Path(docs_dir).rglob("*.qmd")
        quarto_docs = [str(x) for x in quarto_docs]
        quarto_docs = self._filter_ignores(quarto_docs)

        self.dir_context = DirWatcherContext(
            docs_dir, exit_action=self.exit_action, update_on_exit=False
        )
        self.dir_context.enter()
        if quarto_docs:
            for x in quarto_docs:
                x = Path(x)
                parent_path = x.parent
                expected_out = Path(parent_path) / (x.stem + ".md")
                if expected_out.exists():
                    qmd_mtime = x.stat().st_mtime
                    md_mtime = expected_out.stat().st_mtime
                    if qmd_mtime < md_mtime:
                        log.info(f"Skipping {x} as it is older than {expected_out}")
                        continue
                log.info(f"Rendering {x}")
                subprocess.call([quarto, "render", str(x), "--to=markdown"])
        else:
            warnings.warn(f"No quarto files were found in directory {docs_dir}")

        log.info(self.dir_context.update())

    def on_post_build(self, config):
        log.info("Cleaning up:")
        self.dir_context.exit(update=False)



from pathlib import Path
from shutil import copytree
from tempfile import TemporaryDirectory

from .logging import get_logger

log = get_logger(__name__)


class DirWatcherContext:
    """Context manager to watch a directory for changes.

    It checks the context for changes and returns a list of new files.
    When used as a context manager, it will apply the exit action function
    to each new file.

    Example:
    >>> exit_fun = lambda x: print(f"Deleting {x}")
    >>> with DirWatcherContext("dir", exit_action = exit_fun) as ctx:
    ...     # make file A
    ...     new_files = ctx.newfiles
    ...     # make file B
    ... # file a is deleted, file B is kept
    """

    def __init__(self, path, exit_action=None, update_on_exit=True):
        self.logger = get_logger(__name__)
        self.path = path
        self.exit_action = exit_action
        self.newfiles = []
        self.update_on_exit = update_on_exit

    def __enter__(self):
        self.enter()

    def enter(self):
        self.pre_files = list(Path(self.path).rglob("*"))

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.exit(update=self.update_on_exit)

    def exit(self, update=True):
        gen_files = self.newfiles
        if update:
            gen_files = self.update()

        for file in gen_files:
            self.logger.debug(file)
            if self.exit_action:
                self.exit_action(file)
        return gen_files

    def update(self):
        gen_files = list(
            x for x in Path(self.path).rglob("*") if x not in self.pre_files
        )[::-1]
        self.newfiles = gen_files
        return gen_files


class CloneDir(TemporaryDirectory):
    """Clones a directory to a temporary directory.

    It is meant to be used as a context manager.

    Examample:
        >>> with CloneDir("path/to/dir") as clone:
        >>>     print(clone.name)
        >>>     # do stuff
        >>> # clone is deleted
    """

    def __init__(
        self,
        input_dir=Path,
        suffix=None,
        prefix=None,
        out_dir=None,
        ignore_cleanup_errors=False,
    ):
        self.path = Path(input_dir)
        if not self.path.is_dir():
            raise NotADirectoryError(f"{self.path} is not a directory")
        super().__init__(
            suffix=suffix,
            prefix=prefix,
            dir=out_dir,
            ignore_cleanup_errors=ignore_cleanup_errors,
        )
        copytree(str(self.path), self.name, dirs_exist_ok=True)

    def __enter__(self):
        log.debug(f"Cloning {self.path} to {self.name}")
        return super().__enter__()

    def cleanup(self):
        log.debug(f"Cleaning up {self.name}")
        contents = list(Path(self.name).glob("*"))
        contents = [str(x) for x in contents]
        log.debug(f"Removing: {contents}")
        super().cleanup()

    def list_files(self):
        return list(Path(self.name).rglob("*"))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mkquartodocs']

package_data = \
{'': ['*']}

install_requires = \
['jupyter>=1.0.0,<2.0.0',
 'mkdocs>=1.3.1,<2.0.0',
 'nbformat>=5.5.0,<6.0.0',
 'pytest-cov>=4.0.0,<5.0.0']

entry_points = \
{'mkdocs.plugins': ['mkquartodocs = mkquartodocs.plugin:MkDocstringPlugin']}

setup_kwargs = {
    'name': 'mkquartodocs',
    'version': '0.5.1',
    'description': '',
    'long_description': '\n# mkquartodocs\n\n![Example](readme_assets/gif.gif "Example")\n\n**Make gorgeous reproducible documentation with quarto and mkdocs**\n\nIt is a plugin for [mkdocs](https://www.mkdocs.org/) that renders\n[quarto](https://quarto.org) markdown documents into github, so they are\nbuilt with the rest of the documentation.\n\n### Why?\n\nIn many instances the documentation contains runnable code, and it makes\nsense that you verify that the code runs and keep the output of the code\nin sync with the current status of the document and software packages\ninvolved.\n\n**But mainly** I really got tired of manually rendering documents and\ncopying outpus.\n\n## Usage\n\n1.  Install the dependencies: [Installation](#installation)\n2.  Add the plugin to your configuration:\n    [Configuration](#configuration)\n3.  Add `.qmd` files to your `./docs/` directoy\n4.  Run `mkdocs build`\n\nThis will render code chunks and save the outputs! Check out\nhttps://quarto.org/ for more examples on how to use the format.\n\nThis ….\n\n```` markdown\n\n```{python}\nprint(1+1)\n```\n````\n\nWill become this …\n\n``` python\nprint(1+1)\n```\n\n    2\n\n## Installation\n\n1.  Make sure you have quarto installed in your computer.\n\n    - https://quarto.org/docs/get-started/\n\n2.  Install `mkquartodocs`\n\n``` shell\npip install mkquartodocs\n```\n\n## Configuration\n\nAdd `mkquartodocs` to your plugins in your `mkdocs.yml`\n\n``` yaml\n# Whatever is in your mkdocs.yml configuration file....\n# ...\n\nplugins:\n  - mkquartodocs\n```\n\nAvailable configuration options:\n\n- **quarto_path**: Specifies where to look for the quarto executable.\n- **keep_output**: If true it will skip the cleanup step in the\n  directory.\n- **ignore**: a python regular expressions that if matched will mark the\n  file to not be rendered. Note that they need to be full matches\n\n``` yaml\n# Whatever is in your mkdocs.yml configuration file....\n# ...\n\nplugins:\n  - mkquartodocs:\n      quarto_path: /home/my_folder/some/weird/place/to/have/executables/quarto\n      keep_output: true\n      ignore: (.*broken.*.qmd)|(.*page[0-9].qmd)\n```\n\n## Running\n\n**NOTHING !!! you do not have to run it manually!!**\n\nWhen you call `mkdocs build`, it should automatically find your `.qmd`\nfiles, render them, generate the output and clean after itself.\n\n# TODO\n\nThe things that need to/could be added to the project:\n\n- ☐ quarto project support\n- ☐ render in temporary directory, posibly with a ‘safe’ argument\n- ☐ addition of files not in the docs directory, ‘include’ argument\n- ☐ add readme to testing data\n',
    'author': 'J. Sebastian Paez',
    'author_email': 'jspaezp@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
